# Django Imports
from django.contrib.auth import get_user_model

# SOTP Imports
from sotp.models import UserSOTP

# Third Party Imports
from django_apscheduler.models import DjangoJob


def remove_user_otp_and_job(**kwargs):
    """
    It removes the OTP from the database 
    for the user with the email address passed 
    in as a keyword argument
    
    :return: A boolean value.
    """
    user_sotp = UserSOTP.objects.get(
        user=get_user_model().objects.get(
            email=kwargs.get("email")
        )
    )
    user_sotp.totp = 0
    user_sotp.otp = 0
    user_sotp.save()
    
    # Removes user job from database
    DjangoJob.objects.get(id=kwargs.get("job_id")).delete()
    
    return True





# Generated by Django 4.0.4 on 2022-06-04 07:22

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sotp', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='usersotp',
            name='verified',
            field=models.BooleanField(default=False),
        ),
    ]



from django.db import models
from django.contrib.auth import get_user_model


User = get_user_model()


class UserSOTP(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    totp = models.CharField(max_length=18, unique=True, null=True, blank=True)
    otp = models.IntegerField(null=True, blank=True)
    verified = models.BooleanField(default=False)
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=False, null=True, blank=True)
    
    def __str__(self):
        return "User: {} -> TOTP: {}, OTP: {}".format(self.user.username, self.totp, self.otp)
    
    class Meta:
        verbose_name_plural = "User SOTPs"
        db_table = "secured_otps"


# Native Imports Imports
import pyotp

# Django Imports
from django.conf import settings
from django.contrib.auth import get_user_model

# Apps Imports
from sotp.helpers.schedulers import run_scheduler
from sotp.models import UserSOTP


class GenerateSOTP:
    """
    This class generates a random base32 string 
    and uses it to generate a one time password
    """
    
    @classmethod
    def generate_otp(self, user_email:str):
        # Generate a random string of 32 characters
        secret = pyotp.random_base32()    
        
        # Parse the secret to an OTP and set an interval of 900 seconds 
        totp = pyotp.TOTP(secret, interval=settings.SOTP_TIME_EXPIRATION * 60)
        
        # Generate the OTP
        OTP = totp.now()
        
        # Secure TOTP and OTP payload
        payload = {
            "totp": secret,
            "OTP": OTP
        }
        
        # Add generated TOTP and OTP to user and save to database
        user_sotp, _ = UserSOTP.objects.get_or_create(
            user=get_user_model().objects.get(email=user_email)
        )
        user_sotp.totp = payload.get("totp")
        user_sotp.otp = payload.get("OTP")
        user_sotp.save(update_fields=["totp", "otp"])
        
        # Run scheduler to clear user *OTP after interval has elapsed
        run_scheduler(user_email=user_email)
        
        return payload
    



# Generated by Django 4.0.4 on 2022-05-26 20:32

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='UserSOTP',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('totp', models.CharField(blank=True, max_length=18, null=True, unique=True)),
                ('otp', models.IntegerField(blank=True, null=True)),
                ('date_created', models.DateTimeField(auto_now_add=True)),
                ('date_modified', models.DateTimeField(blank=True, null=True)),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name_plural': 'User SOTPs',
                'db_table': 'user_sotps',
            },
        ),
    ]



# Generated by Django 4.0.4 on 2022-10-01 05:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('sotp', '0002_usersotp_verified'),
    ]

    operations = [
        migrations.AlterModelTable(
            name='usersotp',
            table='secured_otps',
        ),
    ]



from django.urls import path
from example.views import redirect_to_register_page, register_page, \
    login_page, confirm_otp_page, welcome_page, logout_user


app_name = "example"


urlpatterns = [
    path("", redirect_to_register_page, name="redirect_to_register_page"),
    path("register/", register_page, name="register-page"),
    path("login/", login_page, name="login-page"),
    path("confirm-otp/", confirm_otp_page, name="confirm-otp-page"),
    path("welcome/", welcome_page, name="welcome-page"),
    path("logout/", logout_user, name="logout_user")
]


# ApScheduler Imports
from apscheduler.schedulers.background import BackgroundScheduler
from django_apscheduler.jobstores import DjangoJobStore, register_events

# Native Imports
import sys
import random

# Django Imports
from django.conf import settings

# SOTP Helpers Imports
from sotp.helpers.remove_otps import remove_user_otp_and_job


def run_scheduler(user_email:str):
    
    # Set scheduler
    scheduler = BackgroundScheduler()
    scheduler.add_jobstore(DjangoJobStore(), "default")
    
    # Register event
    register_events(scheduler)
    
    # Generate Job ID
    job_id = "{}-{}_remove_elapsed_otps"\
            .format(user_email, random.randint(00000000, 9000000))
    
    # run remove user totps and otps job every 15 minutes
    scheduler.add_job(
        remove_user_otp_and_job, 
        'interval', minutes=settings.SOTP_TIME_EXPIRATION,
        jobstore='default',
        id=job_id, max_instances=1,
        replace_existing=True,
        kwargs={"email": user_email, "job_id": job_id}
    )

    # Start scheduler
    scheduler.start() 
    print("Scheduler started...", file=sys.stdout)


import os
from pathlib import Path
from setuptools import setup, find_packages
 

README=Path("README.md").read_text(encoding="utf-8")
 
# Allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))
 
setup(
    name = 'django_sotp',
    version = '1.0.4',
    packages = find_packages(),
    include_package_data = True,
    license = 'CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
    description = 'Generate a secured base32 one time password to authenticate your user! 🔐',
    long_description = README,
    long_description_content_type='text/markdown',
    keywords=[
        'sotp', 
        'otp', 
        'totps', 
        'one time password',
        'passwords' 
    ],
    url = 'https://github.com/israelabraham/django-sopt',
    author = 'Abram',
    author_email = 'israelvictory87@gmail.com',
    classifiers =[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Framework :: Django :: 4.0',
        'Intended Audience :: Developers',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Software Development :: Build Tools',
    ],
    python_requires=">=3.6",
    install_requires=[
        "django>=2.2",
        "django-apscheduler",
        "pyotp",
    ],
)



# Native Imports 
from typing import Tuple
import pyotp

# Django Imports
from django.test import TestCase
from django.conf import settings
from django.contrib.auth import get_user_model as User

# Own Imports
from sotp.models import UserSOTP


class GenerateSOTPTestCase(TestCase):
    """
    Test case to generate a secured one-time password
    """
    
    def setUp(self) -> None:
        self.user = User().objects.get_or_create(
            first_name="Abram",
            last_name="Israel",
            email="abram@email.com",
            username="abram",
            password="someawfully-strong_password2022"
        )[0]
        self.user_sotp =  UserSOTP.objects.get_or_create(
            user=self.user
        )[0]
        
    @property
    def generate_otp(self) -> Tuple[str, str]:
        # Generate a random string of 32 characters
        secret = pyotp.random_base32()    
        
        # Parse the secret to an OTP and set an interval of 900 seconds 
        totp = pyotp.TOTP(secret, interval=settings.SOTP_TIME_EXPIRATION * 60)
        
        # Generate the OTP
        OTP = totp.now()
        
        # Secure TOTP and OTP payload
        payload = {
            "totp": secret,
            "OTP": OTP
        }
        return payload["totp"], payload["OTP"]
    
    def test_generate_sotp_for_user(self):
        """
        Test case to ensure that the generated
        TOTP and OTP matches with the user's.
        """
        
        totp, otp = self.generate_otp
        
        user = self.user_sotp
        user.otp = otp
        user.totp = totp
        user.save(update_fields=["otp", "totp"])
        
        self.assertEqual(user.otp, otp)
        self.assertEqual(user.totp, totp)


"""
Django settings for core project.

Generated by 'django-admin startproject' using Django 4.0.4.

For more information on this file, see
https://docs.djangoproject.com/en/4.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/4.0/ref/settings/
"""

from django.contrib.messages import constants as messages
from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/4.0/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'add-something-here'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    
    'example',
    'sotp',
    
    'django_apscheduler',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'core.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'core.wsgi.application'


# Database
# https://docs.djangoproject.com/en/4.0/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': BASE_DIR / 'db.sqlite3',
    }
}


# Password validation
# https://docs.djangoproject.com/en/4.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
# https://docs.djangoproject.com/en/4.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'Africa/Lagos'

USE_I18N = True

USE_TZ = False


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/4.0/howto/static-files/

STATIC_URL = 'static/'


# Message Configuration
MESSAGE_TAGS = {
    messages.SUCCESS: 'alert alert-success',
    messages.WARNING: 'alert alert-warning',
    messages.ERROR: 'alert alert-danger',
}

# Default primary key field type
# https://docs.djangoproject.com/en/4.0/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

# Format string for displaying run time timestamps in the Django admin site. The default
# just adds seconds to the standard Django format, which is useful for displaying the timestamps
# for jobs that are scheduled to run on intervals of less than one minute.
# 
# See https://docs.djangoproject.com/en/dev/ref/settings/#datetime-format for format string
# syntax details.
APSCHEDULER_DATETIME_FORMAT = "N j, Y, f:s a"

# Maximum run time allowed for jobs that are triggered manually via the Django admin site, which
# prevents admin site HTTP requests from timing out.
# 
# Longer running jobs should probably be handed over to a background task processing library
# that supports multiple background worker processes instead (e.g. Dramatiq, Celery, Django-RQ,
# etc. See: https://djangopackages.org/grids/g/workers-queues-tasks/ for popular options).
APSCHEDULER_RUN_NOW_TIMEOUT = 25  # Second

# SOTP configuration    
SOTP_TIME_EXPIRATION = 5 # in minutes